from __future__ import annotations

import contextlib
import json
import os
import pickle
import subprocess
import tempfile
import warnings
import zipfile
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, TextIO, Union

import fsspec

from rubicon_ml import client, domain
from rubicon_ml.client.utils.exception_handling import failsafe
from rubicon_ml.client.utils.tags import TagContainer, filter_children
from rubicon_ml.domain import Artifact as ArtifactDomain
from rubicon_ml.exceptions import RubiconException

if TYPE_CHECKING:
    import dask.dataframe as dd
    import pandas as pd
    import polars as pl
    import xgboost as xgb

    from rubicon_ml.client import Artifact, Dataframe
    from rubicon_ml.domain import DOMAIN_TYPES


class ArtifactMixin:
    """Adds artifact support to a client object."""

    _domain: ArtifactDomain

    def _validate_data(self, data_bytes, data_directory, data_file, data_object, data_path, name):
        """Raises a `RubiconException` if the data to log as an artifact is improperly provided."""
        if not any([data_bytes, data_directory, data_file, data_object, data_path]):
            raise RubiconException(
                "One of `data_bytes`, `data_directory`, `data_file`, `data_object` or "
                "`data_path` must be provided."
            )

        if name is None:
            if data_path is not None:
                name = os.path.basename(data_path)
            else:
                raise RubiconException("`name` must be provided if not using `data_path`.")

        if data_directory is not None:
            temp_file_context = tempfile.TemporaryDirectory
        else:
            temp_file_context = contextlib.nullcontext

        if data_bytes is None:
            with temp_file_context() as temp_dir:
                if data_object is not None:
                    data_bytes = pickle.dumps(data_object)
                else:
                    if data_directory is not None:
                        temp_zip_name = Path(f"{temp_dir}/{name}")

                        with zipfile.ZipFile(str(temp_zip_name), "w") as zip_file:
                            for dir_path, _, files in os.walk(data_directory):
                                for file in files:
                                    zip_file.write(Path(f"{dir_path}/{file}"), arcname=file)

                        file = fsspec.open(temp_zip_name, "rb")
                    elif data_file is not None:
                        file = data_file
                    elif data_path is not None:
                        file = fsspec.open(data_path, "rb")

                    with file as open_file:
                        data_bytes = open_file.read()

        return data_bytes, name

    @failsafe
    def log_artifact(
        self,
        data_bytes: Optional[bytes] = None,
        data_directory: Optional[str] = None,
        data_file: Optional[TextIO] = None,
        data_object: Optional[Any] = None,
        data_path: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        comments: Optional[List[str]] = None,
    ) -> Artifact:
        """Log an artifact to this client object.

        Parameters
        ----------
        data_bytes : bytes, optional
            The raw bytes to log as an artifact.
        data_directory : str, optional
            The path to a directory to zip and log as an artifact.
        data_file : TextIOWrapper, optional
            The open file to log as an artifact.
        data_object : python object, optional
            The python object to log as an artifact.
        data_path : str, optional
            The absolute or relative local path or S3 path
            to the data to log as an artifact. S3 paths
            must be prepended with 's3://'.
        name : str, optional
            The name of the artifact file. Required if
            `data_path` is not provided.
        description : str, optional
            A description of the artifact. Use to provide
            additional context.
        tags : list of str, optional
            Values to tag the experiment with. Use tags to organize and
            filter your artifacts.
        comments : list of str, optional
            Values to comment the experiment with. Use comments to organize and
            filter your artifacts.

        Notes
        -----
        Only one of `data_bytes`, `data_file`, `data_object`, and `data_path`
        should be provided. If more than one is given, the order
        of precedence is `data_bytes`, `data_object`, `data_file`, `data_path`.

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.

        Examples
        --------
        >>> # Log with bytes
        >>> experiment.log_artifact(
        ...     data_bytes=b'hello rubicon!',
        ...     name="bytes_artifact",
        ...     description="log artifact from bytes",
        ... )

        >>> # Log zipped directory
        >>> experiment.log_artifact(
        ...     data_directory="./path/to/directory/",
        ...     name="directory.zip",
        ...     description="log artifact from zipped directory",
        ... )

        >>> # Log with file
        >>> with open('./path/to/artifact.txt', 'rb') as file:
        >>>     project.log_artifact(
        ...         data_file=file,
        ...         name="file_artifact",
        ...         description="log artifact from file",
        ...     )

        >>> # Log with file path
        >>> experiment.log_artifact(
        ...     data_path="./path/to/artifact.pkl",
        ...     description="log artifact from file path",
        ... )
        """
        if tags is None:
            tags = []
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        if comments is None:
            comments = []
        if not isinstance(comments, list) or not all(
            [isinstance(comment, str) for comment in comments]
        ):
            raise ValueError("`comments` must be `list` of type `str`")

        data_bytes, name = self._validate_data(
            data_bytes, data_directory, data_file, data_object, data_path, name
        )

        artifact = domain.Artifact(
            name=name,
            description=description,
            parent_id=self._domain.id,
            tags=tags,
            comments=comments,
        )

        project_name, experiment_id = self._get_identifiers()
        for repo in self.repositories:
            repo.create_artifact(artifact, data_bytes, project_name, experiment_id=experiment_id)

        return client.Artifact(artifact, self)

    def _get_environment_bytes(self, export_cmd: List[str]) -> bytes:
        """Get the working environment as a sequence of bytes.

        Parameters
        ----------
        export_cmd : list of str
            The command to export the environment.

        Returns
        -------
        bytes
            A bytes sequence of the environment.
        """
        try:
            completed_process = subprocess.run(export_cmd, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            raise RubiconException(e.stderr)

        return completed_process.stdout

    @failsafe
    def log_conda_environment(self, artifact_name: Optional[str] = None) -> Artifact:
        """Log the conda environment as an artifact to this client object.
        Useful for recreating your exact environment at a later date.

        Parameters
        ----------
        artifact_name : str, optional
            The name of the artifact (the exported conda environment).

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.

        Notes
        -----
        Relies on running with an active conda environment.
        """
        if artifact_name is None:
            artifact_name = f"environment-{datetime.now().strftime('%Y_%m_%d-%I_%M_%S_%p')}.yml"

        env_bytes = self._get_environment_bytes(["conda", "env", "export"])
        artifact = self.log_artifact(data_bytes=env_bytes, name=artifact_name)

        return artifact

    @failsafe
    def log_h2o_model(
        self,
        h2o_model,
        artifact_name: Optional[str] = None,
        export_cross_validation_predictions: bool = False,
        use_mojo: bool = False,
        **log_artifact_kwargs,
    ) -> Artifact:
        """Log an `h2o` model as an artifact using `h2o.save_model`.

        Parameters
        ----------
        h2o_model : h2o.model.ModelBase
            The `h2o` model to log as an artifact.
        artifact_name : str, optional (default None)
            The name of the artifact. Defaults to None, using `h2o_model`'s class name.
        export_cross_validation_predictions: bool, optional (default False)
            Passed directly to `h2o.save_model`.
        use_mojo: bool, optional (default False)
            Whether to log the model in MOJO format. If False, the model will be
            logged in binary format.
        log_artifact_kwargs : dict
            Additional kwargs to be passed directly to `self.log_artifact`.
        """
        import h2o

        if self.repository.PROTOCOL == "memory":
            raise RubiconException("`h2o` models cannot be logged in memory with `log_h2o_model`.")

        if artifact_name is None:
            artifact_name = h2o_model.__class__.__name__

        with tempfile.TemporaryDirectory() as temp_dir_name:
            if use_mojo:
                model_data_path = f"{temp_dir_name}/{artifact_name}.zip"
                h2o_model.download_mojo(path=model_data_path)
            else:
                model_data_path = h2o.save_model(
                    h2o_model,
                    export_cross_validation_predictions=export_cross_validation_predictions,
                    filename=artifact_name,
                    path=temp_dir_name,
                )

            artifact = self.log_artifact(
                name=artifact_name,
                data_path=model_data_path,
                **log_artifact_kwargs,
            )

        return artifact

    @failsafe
    def log_xgboost_model(
        self,
        xgboost_model: "xgb.Booster",
        artifact_name: Optional[str] = None,
        **log_artifact_kwargs: Any,
    ) -> Artifact:
        """Log an XGBoost model as a JSON file to this client object.

        Please note that we do not currently support logging directly from the SKLearn interface.

        Parameters
        ----------
        xgboost_model: Booster
            An xgboost model object in the Booster format
        artifact_name : str, optional
            The name of the artifact (the exported XGBoost model).
        log_artifact_kwargs : Any
            Additional kwargs to be passed directly to `self.log_artifact`.

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.
        """
        if artifact_name is None:
            artifact_name = xgboost_model.__class__.__name__

        # TODO: handle sklearn
        booster = xgboost_model

        with tempfile.TemporaryDirectory() as temp_dir_name:
            model_location = f"{temp_dir_name}/{artifact_name}.json"
            booster.save_model(model_location)

            artifact = self.log_artifact(
                name=artifact_name,
                data_path=model_location,
                **log_artifact_kwargs,
            )

        return artifact

    @failsafe
    def log_pip_requirements(self, artifact_name: Optional[str] = None) -> Artifact:
        """Log the pip requirements as an artifact to this client object.
        Useful for recreating your exact environment at a later date.

        Parameters
        ----------
        artifact_name : str, optional
            The name of the artifact (the exported pip environment).

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.
        """
        if artifact_name is None:
            artifact_name = f"requirements-{datetime.now().strftime('%Y_%m_%d-%I_%M_%S_%p')}.txt"

        requirements_bytes = self._get_environment_bytes(["pip", "freeze"])
        artifact = self.log_artifact(data_bytes=requirements_bytes, name=artifact_name)

        return artifact

    @failsafe
    def artifacts(
        self,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
        qtype: str = "or",
    ) -> List[Artifact]:
        """Get the artifacts logged to this client object.

        Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Artifact
            The artifacts previously logged to this client object.
        """
        if tags is None:
            tags = []
        project_name, experiment_id = self._get_identifiers()
        return_err = None
        for repo in self.repositories:
            try:
                artifacts = [
                    client.Artifact(a, self)
                    for a in repo.get_artifacts_metadata(project_name, experiment_id=experiment_id)
                ]
            except Exception as err:
                return_err = err
            else:
                self._artifacts = filter_children(artifacts, tags, qtype, name)
                return self._artifacts

        self._raise_rubicon_exception(return_err)

    @failsafe
    def artifact(self, name: Optional[str] = None, id: Optional[str] = None) -> Artifact:
        """Get an artifact logged to this project by id or name.

        Parameters
        ----------
        id : str
            The id of the artifact to get.
        name : str
            The name of the artifact to get.

        Returns
        -------
        rubicon.client.Artifact
            The artifact logged to this project with id `id` or name 'name'.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            artifacts = self.artifacts(name=name)

            if len(artifacts) == 0:
                raise RubiconException(f"No artifact found with name '{name}'.")
            if len(artifacts) > 1:
                warnings.warn(
                    f"Multiple artifacts found with name '{name}'. Returning most recently logged."
                )

            artifact = artifacts[-1]
            return artifact
        else:
            project_name, experiment_id = self._get_identifiers()
            return_err = None
            for repo in self.repositories:
                try:
                    artifact = client.Artifact(
                        repo.get_artifact_metadata(project_name, id, experiment_id),
                        self,
                    )
                except Exception as err:
                    return_err = err
                else:
                    return artifact

        self._raise_rubicon_exception(return_err)

    @failsafe
    def delete_artifacts(self, ids: List[str]):
        """Delete the artifacts logged to with client object
        with ids `ids`.

        Parameters
        ----------
        ids : list of str
            The ids of the artifacts to delete.
        """
        project_name, experiment_id = self._get_identifiers()

        for artifact_id in ids:
            for repo in self.repositories:
                repo.delete_artifact(project_name, artifact_id, experiment_id=experiment_id)

    @failsafe
    def log_json(
        self,
        json_object: Dict[str, Any],
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
    ) -> Artifact:
        """Log a python dictionary to a JSON file.

        Parameters
        ----------
        json_object : Dict[str, Any]
            A python dictionary capable of being converted to JSON.
        name : Optional[str], optional
            A name for this JSON file, by default None
        description : Optional[str], optional
            A description for this file, by default None
        tags : Optional[List[str]], optional
            Any Rubicon tags, by default None

        Returns
        -------
        Artifact
            The new artifact.

        """
        if name is None:
            json_name = f"dictionary-{datetime.now().strftime('%Y_%m_%d-%I_%M_%S_%p')}.json"
        else:
            json_name = name

        artifact = self.log_artifact(
            data_bytes=bytes(json.dumps(json_object), "utf-8"),
            name=json_name,
            description=description,
            tags=tags,
        )

        return artifact


class DataframeMixin:
    """Adds dataframe support to a client object."""

    _domain: DOMAIN_TYPES

    @failsafe
    def log_dataframe(
        self,
        df: Union[pd.DataFrame, "dd.DataFrame", "pl.DataFrame"],
        description: Optional[str] = None,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
        comments: Optional[List[str]] = None,
    ) -> Dataframe:
        """Log a dataframe to this client object.

        Parameters
        ----------
        df : pandas.DataFrame, dask.dataframe.DataFrame, or polars DataFrame
            The dataframe to log.
        description : str, optional
            The dataframe's description. Use to provide
            additional context.
        tags : list of str
            The values to tag the dataframe with.
        comments: list of str
            The values to comment the dataframe with.

        Returns
        -------
        rubicon.client.Dataframe
            The new dataframe.
        """
        if tags is None:
            tags = []
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        if comments is None:
            comments = []
        if not isinstance(comments, list) or not all(
            [isinstance(comment, str) for comment in comments]
        ):
            raise ValueError("`comments` must be `list` of type `str`")

        dataframe = domain.Dataframe(
            parent_id=self._domain.id,
            description=description,
            name=name,
            tags=tags,
            comments=comments,
        )

        project_name, experiment_id = self._get_identifiers()
        for repo in self.repositories:
            repo.create_dataframe(dataframe, df, project_name, experiment_id=experiment_id)

        return client.Dataframe(dataframe, self)

    @failsafe
    def dataframes(
        self,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
        qtype: str = "or",
    ) -> List[Dataframe]:
        """Get the dataframes logged to this client object.

        Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Dataframe
            The dataframes previously logged to this client object.
        """
        if tags is None:
            tags = []
        project_name, experiment_id = self._get_identifiers()
        return_err = None
        for repo in self.repositories:
            try:
                dataframes = [
                    client.Dataframe(d, self)
                    for d in repo.get_dataframes_metadata(project_name, experiment_id=experiment_id)
                ]
            except Exception as err:
                return_err = err
            else:
                self._dataframes = filter_children(dataframes, tags, qtype, name)
                return self._dataframes

        self._raise_rubicon_exception(return_err)

    @failsafe
    def dataframe(self, name: Optional[str] = None, id: Optional[str] = None) -> Dataframe:
        """
        Get the dataframe logged to this client object.

        Parameters
        ----------
        id : str
            The id of the dataframe to get.
        name : str
            The name of the dataframe to get.
        Returns
        -------
        rubicon.client.Dataframe
            The dataframe logged to this project with id `id` or name 'name'.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        elif name is not None:
            dataframes = self.dataframes(name=name)

            if len(dataframes) == 0:
                raise RubiconException(f"No dataframe found with name '{name}'.")
            elif len(dataframes) > 1:
                warnings.warn(
                    f"Multiple dataframes found with name '{name}'. Returning most recently logged."
                )

            dataframe = dataframes[-1]
            return dataframe
        else:
            project_name, experiment_id = self._get_identifiers()
            return_err = None
            for repo in self.repositories:
                try:
                    dataframe = client.Dataframe(
                        repo.get_dataframe_metadata(
                            project_name, experiment_id=experiment_id, dataframe_id=id
                        ),
                        self,
                    )
                except Exception as err:
                    return_err = err
                else:
                    return dataframe

        self._raise_rubicon_exception(return_err)

    @failsafe
    def delete_dataframes(self, ids: List[str]):
        """Delete the dataframes with ids `ids` logged to
        this client object.

        Parameters
        ----------
        ids : list of str
            The ids of the dataframes to delete.
        """
        project_name, experiment_id = self._get_identifiers()

        for dataframe_id in ids:
            for repo in self.repositories:
                repo.delete_dataframe(project_name, dataframe_id, experiment_id=experiment_id)


class TagMixin:
    """Adds tag support to a client object."""

    _domain: DOMAIN_TYPES

    def _get_taggable_identifiers(self):
        project_name, experiment_id = self._parent._get_identifiers()
        entity_identifier = None

        # experiments do not return an entity identifier - they are the entity
        if isinstance(self, client.Experiment):
            experiment_id = self.id
        # dataframes and artifacts are identified by their `id`s
        elif isinstance(self, client.Dataframe) or isinstance(self, client.Artifact):
            entity_identifier = self.id
        # everything else is identified by its `name`
        else:
            entity_identifier = self.name

        return project_name, experiment_id, entity_identifier

    @failsafe
    def add_tags(self, tags: List[str]):
        """Add tags to this client object.

        Parameters
        ----------
        tags : list of str
            The tag values to add.
        """
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        project_name, experiment_id, entity_identifier = self._get_taggable_identifiers()

        self._domain.add_tags(tags)
        for repo in self.repositories:
            repo.add_tags(
                project_name,
                tags,
                experiment_id=experiment_id,
                entity_identifier=entity_identifier,
                entity_type=self.__class__.__name__,
            )

    @failsafe
    def remove_tags(self, tags: List[str]):
        """Remove tags from this client object.

        Parameters
        ----------
        tags : list of str
             The tag values to remove.
        """
        project_name, experiment_id, entity_identifier = self._get_taggable_identifiers()

        self._domain.remove_tags(tags)
        for repo in self.repositories:
            repo.remove_tags(
                project_name,
                tags,
                experiment_id=experiment_id,
                entity_identifier=entity_identifier,
                entity_type=self.__class__.__name__,
            )

    def _update_tags(self, tag_data):
        """Add or remove the tags in `tag_data` based on
        their key.
        """
        for tag in tag_data:
            self._domain.add_tags(tag.get("added_tags", []))
            self._domain.remove_tags(tag.get("removed_tags", []))

    @property
    def tags(self) -> TagContainer:
        """Get this client object's tags."""
        project_name, experiment_id, entity_identifier = self._get_taggable_identifiers()
        return_err = None
        for repo in self.repositories:
            try:
                tag_data = repo.get_tags(
                    project_name,
                    experiment_id=experiment_id,
                    entity_identifier=entity_identifier,
                    entity_type=self.__class__.__name__,
                )
            except Exception as err:
                return_err = err
            else:
                self._update_tags(tag_data)

                return TagContainer(self._domain.tags)

        self._raise_rubicon_exception(return_err)


class CommentMixin:
    """Adds comment support to a client object."""

    _domain: DOMAIN_TYPES

    def _get_commentable_identifiers(self):
        project_name, experiment_id = self._parent._get_identifiers()
        entity_identifier = None

        # experiments do not return an entity identifier - they are the entity
        if isinstance(self, client.Experiment):
            experiment_id = self.id
        # dataframes and artifacts are identified by their `id`s
        elif isinstance(self, client.Dataframe) or isinstance(self, client.Artifact):
            entity_identifier = self.id
        # everything else is identified by its `name`
        else:
            entity_identifier = self.name

        return project_name, experiment_id, entity_identifier

    @failsafe
    def add_comments(self, comments: List[str]):
        """Add comments to this client object.

        Parameters
        ----------
        comments : list of str
            The comment values to add.
        """
        if not isinstance(comments, list) or not all(
            [isinstance(comment, str) for comment in comments]
        ):
            raise ValueError("`comments` must be `list` of type `str`")

        project_name, experiment_id, entity_identifier = self._get_commentable_identifiers()

        self._domain.add_comments(comments)
        for repo in self.repositories:
            repo.add_comments(
                project_name,
                comments,
                experiment_id=experiment_id,
                entity_identifier=entity_identifier,
                entity_type=self.__class__.__name__,
            )

    @failsafe
    def remove_comments(self, comments: List[str]):
        """Remove comments from this client object.

        Parameters
        ----------
        comments : list of str
             The comment values to remove.
        """
        project_name, experiment_id, entity_identifier = self._get_commentable_identifiers()

        self._domain.remove_comments(comments)
        for repo in self.repositories:
            repo.remove_comments(
                project_name,
                comments,
                experiment_id=experiment_id,
                entity_identifier=entity_identifier,
                entity_type=self.__class__.__name__,
            )

    def _update_comments(self, comment_data):
        """Add or remove the comments in `comment_data` based on
        their key.
        """
        for comment in comment_data:
            self._domain.add_comments(comment.get("added_comments", []))
            self._domain.remove_comments(comment.get("removed_comments", []))

    @property
    def comments(self) -> List[str]:
        """Get this client object's comments."""
        project_name, experiment_id, entity_identifier = self._get_commentable_identifiers()
        return_err = None
        for repo in self.repositories:
            try:
                comment_data = repo.get_comments(
                    project_name,
                    experiment_id=experiment_id,
                    entity_identifier=entity_identifier,
                    entity_type=self.__class__.__name__,
                )
            except Exception as err:
                return_err = err
            else:
                self._update_comments(comment_data)

                return self._domain.comments

        self._raise_rubicon_exception(return_err)



import os

import dash_bootstrap_components as dbc
from dash import dash_table, dcc, html
from dash.dependencies import ALL, Input, Output, State

from rubicon_ml.intake_rubicon.publish import publish
from rubicon_ml.viz.base import VizBase
from rubicon_ml.viz.common.colors import light_blue, plot_background_blue


class ExperimentsTable(VizBase):
    """Visualize the experiments `experiments` and their metadata, metrics,
    and parameters in a tabular format.

    Parameters
    ----------
    experiments : list of rubicon_ml.client.experiment.Experiment, optional
        The experiments to visualize. Defaults to None. Can be set as
        attribute after instantiation.
    is_selectable : bool, optional
        True to enable selection of the rows in the table, False otherwise.
        Defaults to True.
    metric_names : list of str
        If provided, only show the metrics with names in the given list. If
        `metric_query_tags` are also provided, this will only select metrics
        from the tag-filtered results.
    metric_query_tags : list of str, optional
        If provided, only show the metrics with the given tags in the table.
    metric_query_type : 'and' or 'or', optional
        When `metric_query_tags` are given, 'and' shows the metrics with all of
        the given tags and 'or' shows the metrics with any of the given tags.
    parameter_names : list of str
        If provided, only show the parameters with names in the given list. If
        `parameter_query_tags` are also provided, this will only select
        parameters from the tag-filtered results.
    parameter_query_tags : list of str, optional
        If provided, only show the parameters with the given tags in the table.
    parameter_query_type : 'and' or 'or', optional
        When `parameter_query_tags` are given, 'and' shows the paramters with
        all of the given tags and 'or' shows the parameters with any of the
        given tags.
    """

    def __init__(
        self,
        experiments=None,
        is_selectable=True,
        metric_names=None,
        metric_query_tags=None,
        metric_query_type=None,
        parameter_names=None,
        parameter_query_tags=None,
        parameter_query_type=None,
    ):
        super().__init__(dash_title="experiment table")

        self.experiments = experiments
        self.is_selectable = is_selectable
        self.metric_names = metric_names
        self.metric_query_tags = metric_query_tags
        self.metric_query_type = metric_query_type
        self.parameter_names = parameter_names
        self.parameter_query_tags = parameter_query_tags
        self.parameter_query_type = parameter_query_type

    @property
    def layout(self):
        """Defines the experiments table's layout."""
        bulk_select_buttons = [
            html.Div(
                dbc.Button(
                    "select all experiments",
                    color="primary",
                    disabled=not self.is_selectable,
                    id="select-all-button",
                    outline=True,
                ),
                className="bulk-select-button-container",
            ),
            html.Div(
                dbc.Button(
                    "clear all experiments",
                    color="primary",
                    disabled=not self.is_selectable,
                    id="clear-all-button",
                    outline=True,
                ),
                className="bulk-select-button-container",
            ),
            html.Div(
                dbc.Button(
                    "publish selected",
                    color="primary",
                    disabled=not self.is_selectable,
                    id="publish-selected-button",
                    outline=True,
                ),
                className="bulk-select-button-container",
            ),
        ]

        experiment_table = dash_table.DataTable(
            columns=[
                {"name": column, "id": column, "selectable": self.is_selectable}
                for column in self.all_columns
            ],
            data=self.experiment_records,
            filter_action="native",
            fixed_columns={"headers": True, "data": 1},
            hidden_columns=self.hidden_columns,
            id="experiment-table",
            page_size=10,
            row_selectable="multi" if self.is_selectable else False,
            selected_rows=[],
            sort_action="native",
            sort_mode="multi",
            style_cell={"overflow": "hidden", "textOverflow": "ellipsis"},
            style_data_conditional=[
                {"if": {"row_index": "odd"}, "backgroundColor": plot_background_blue}
            ],
            style_header={"fontWeight": 700},
            style_table={"minWidth": "100%"},
        )

        header_text = (
            f"showing {len(self.experiments)} experiments "
            f"{'at commit ' if self.commit_hash is not None else ''}"
        )

        header = html.H5(
            [
                html.P(
                    header_text,
                    className="experiment-table-header-text",
                    style={"float": "left"} if self.commit_hash is not None else {},
                ),
                html.A(
                    html.P(
                        [
                            self.commit_hash,
                            html.I(className="bi bi-box-arrow-up-right external-link-icon"),
                        ]
                    ),
                    href=self.github_url,
                    style={"display": "none"} if self.commit_hash is None else {},
                    target="_blank",
                ),
            ],
            className="header-text",
        )

        toggle_columns_dropdown = dbc.DropdownMenu(
            [
                dbc.DropdownMenuItem(
                    "show all",
                    id="show-all-dropdown-button",
                ),
                dbc.DropdownMenuItem(
                    "hide all",
                    id="hide-all-dropdown-button",
                ),
                dbc.DropdownMenuItem(divider=True),
                *[
                    dbc.DropdownMenuItem(
                        dcc.Checklist(
                            inputClassName="column-dropdown-checkbox",
                            labelClassName="column-dropdown-label",
                            options=[{"label": column, "value": column}],
                            value=[column] if column not in self.hidden_columns else [],
                            id={
                                "type": "column-dropdown-checkbox",
                                "index": column,
                            },
                        ),
                        id={"type": "column-dropdown-button", "index": column},
                    )
                    for column in self.all_columns
                ],
            ],
            color="secondary",
            id="column-selection-dropdown",
            label="toggle columns",
        )

        publish_modal = dbc.Modal(
            [
                dbc.ModalHeader(
                    dbc.ModalTitle("publish selected experiments"),
                    close_button=True,
                ),
                dbc.ModalBody(
                    [
                        dbc.Label("enter catalog YAML output path:"),
                        dbc.Input(
                            id="publish-path-input",
                            type="text",
                            value=os.path.join(os.getcwd(), "rubicon-ml-catalog.yml"),
                        ),
                    ],
                ),
                dbc.ModalFooter(dbc.Button("publish", id="publish-button")),
            ],
            id="publish-modal",
            centered=True,
            is_open=False,
            size="lg",
        )

        if self.is_selectable:
            header_row = [
                html.Div(
                    header,
                    className="header-row",
                ),
                dbc.Row(
                    [
                        *[dbc.Col(button, width="auto") for button in bulk_select_buttons],
                        dbc.Col(toggle_columns_dropdown),
                    ],
                    className="button-group",
                ),
            ]
        else:
            header_row = [
                dbc.Row(
                    [
                        dbc.Col(
                            html.Div(
                                header,
                                className="header-row",
                            ),
                            width=8,
                        ),
                        dbc.Col(toggle_columns_dropdown),
                    ],
                    className="button-group",
                ),
            ]

        return html.Div(
            [
                *header_row,
                publish_modal,
                dcc.Loading(experiment_table, color=light_blue),
            ],
        )

    def load_experiment_data(self):
        """Load the experiment data required for the experiments table.

        Extracts all experiment metadata as well as parameters and metrics
        from each experiment in `self.experiments`. Sets GitHub information
        if applicable.
        """
        self.experiment_records = []

        self.all_columns = ["id", "name", "created_at", "model_name", "commit_hash", "tags"]
        self.hidden_columns = []

        self.commit_hash = None
        self.github_url = None

        all_parameter_names = set()
        all_metric_names = set()
        commit_hashes = set()
        show_columns = {"id", "created_at"}

        for experiment in self.experiments:
            experiment_record = {
                "id": experiment.id,
                "name": experiment.name,
                "created_at": experiment.created_at,
                "model_name": experiment.model_name,
                "commit_hash": None,
                "tags": ", ".join(str(tag) for tag in experiment.tags),
            }

            if experiment.commit_hash:
                experiment_record["commit_hash"] = experiment.commit_hash[:7]

                commit_hashes.add(experiment.commit_hash)
                show_columns.add("commit_hash")

            if experiment.model_name is not None:
                show_columns.add("model_name")

            if experiment.name is not None:
                show_columns.add("name")

            if len(experiment.tags) > 0:
                show_columns.add("tags")

            for parameter in experiment.parameters():
                experiment_record[parameter.name] = str(parameter.value)

                all_parameter_names.add(parameter.name)

            for metric in experiment.metrics():
                experiment_record[metric.name] = str(metric.value)

                all_metric_names.add(metric.name)

            self.experiment_records.append(experiment_record)

        if self.parameter_query_tags is not None:
            parameters = experiment.parameters(
                tags=self.parameter_query_tags,
                qtype=self.parameter_query_type,
            )
            show_parameter_names = set([p.name for p in parameters])
        else:
            show_parameter_names = all_parameter_names

        if self.parameter_names is not None:
            show_parameter_names = set(
                [name for name in show_parameter_names if name in self.parameter_names]
            )

        if self.metric_query_tags is not None:
            metrics = experiment.metrics(
                tags=self.metric_query_tags,
                qtype=self.metric_query_type,
            )
            show_metric_names = set([m.name for m in metrics])
        else:
            show_metric_names = all_metric_names

        if self.metric_names is not None:
            show_metric_names = set(
                [name for name in show_metric_names if name in self.metric_names]
            )

        self.all_columns.extend(list(all_parameter_names) + list(all_metric_names))
        self.hidden_columns = [
            column
            for column in self.all_columns
            if column not in show_columns | show_metric_names | show_parameter_names
        ]

        if len(commit_hashes) == 1:
            github_url_root = self.experiments[0].project.github_url[:-4]
            commit_hash = list(commit_hashes)[0]

            self.commit_hash = commit_hash[:7]
            self.github_url = f"{github_url_root}/tree/{commit_hash}"

    def register_callbacks(self, link_experiment_table=False):
        @self.app.callback(
            Output({"type": "column-dropdown-checkbox", "index": ALL}, "value"),
            [
                Input("show-all-dropdown-button", "n_clicks_timestamp"),
                Input("hide-all-dropdown-button", "n_clicks_timestamp"),
            ],
            prevent_initial_call=True,
        )
        def update_selected_column_checkboxes(last_show_click, last_hide_click):
            """Bulk updates for the selections in the "toggle columns" dropdown.

            Returns all if triggered by the "show all" button and returns
            only "id" if triggered by the "hide all" button. The initial
            call is prevented as the default state is neither all nor none.
            """
            last_show_click = last_show_click if last_show_click else 0
            last_hide_click = last_hide_click if last_hide_click else 0

            if last_hide_click > last_show_click:
                hidden_values = [[]] * (len(self.all_columns) - 1)

                return [["id"], *hidden_values]

            return [[column] for column in self.all_columns]

        @self.app.callback(
            Output("experiment-table", "hidden_columns"),
            Input({"type": "column-dropdown-checkbox", "index": ALL}, "value"),
        )
        def update_hidden_experiment_table_cols(selected_columns):
            """Hide and show the columns in the experiment table.

            Returns the columns that should be hidden based on whether or
            not the column's corresponding value is checked in the "toggle
            columns" dropdown.
            """
            selected_columns = [sc[0] for sc in selected_columns if len(sc) > 0]

            return [column for column in self.all_columns if column not in selected_columns]

        @self.app.callback(
            Output("experiment-table", "selected_rows"),
            [
                Input("select-all-button", "n_clicks_timestamp"),
                Input("clear-all-button", "n_clicks_timestamp"),
            ],
            State("experiment-table", "derived_virtual_indices"),
        )
        def update_selected_experiment_table_rows(
            last_select_click, last_clear_click, experiment_table_indices
        ):
            """Bulk selection for the rows of the experiment table.

            Returns all if triggered by the "select all" button and returns
            none if triggered by the "clear all" button.
            """
            if last_select_click is None and last_clear_click is None:
                return list(range(len(self.experiments)))

            last_select_click = last_select_click if last_select_click else 0
            last_clear_click = last_clear_click if last_clear_click else 0

            if last_select_click > last_clear_click:
                return experiment_table_indices

            return []

        @self.app.callback(
            Output("publish-modal", "is_open"),
            [
                Input("publish-selected-button", "n_clicks_timestamp"),
                Input("publish-button", "n_clicks_timestamp"),
            ],
            [
                State("publish-modal", "is_open"),
                State("publish-path-input", "value"),
                State("experiment-table", "derived_virtual_selected_rows"),
                State("experiment-table", "derived_virtual_data"),
            ],
        )
        def toggle_publish_modal(
            last_publish_selected_click,
            last_publish_click,
            is_modal_open,
            publish_path,
            selected_rows,
            data,
        ):
            last_publish_selected_click = (
                last_publish_selected_click if last_publish_selected_click else 0
            )
            last_publish_click = last_publish_click if last_publish_click else 0

            if last_publish_selected_click > last_publish_click:
                return True
            elif last_publish_click > last_publish_selected_click:
                selected_experiment_ids = [data[row].get("id") for row in selected_rows]
                selected_experiments = [
                    e for e in self.experiments if e.id in selected_experiment_ids
                ]

                publish(selected_experiments, publish_path)

                return False

            return is_modal_open



import os
import subprocess
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union

from rubicon_ml import domain
from rubicon_ml.client import Config, Project
from rubicon_ml.client.utils.exception_handling import failsafe
from rubicon_ml.domain.utils import TrainingMetadata
from rubicon_ml.exceptions import RubiconException
from rubicon_ml.repository.utils import slugify


class Rubicon:
    """The `rubicon` client's entry point.

    Creates a `Config` and injects it into the client level
    objects at run-time.

    Parameters
    ----------
    persistence : str, optional
        The persistence type. Can be one of ["filesystem", "memory"].
        Defaults to "filesystem".
    root_dir : str, optional
        Absolute or relative filepath. Use absolute path for best performance.
        Defaults to the local filesystem. Prefix with s3:// to use s3 instead.
    auto_git_enabled : bool, optional
        True to use the `git` command to automatically log relevant repository
        information to projects and experiments logged with this client instance,
        False otherwise. Defaults to False.
    storage_options : dict, optional
        Additional keyword arguments specific to the protocol being chosen. They
        are passed directly to the underlying filesystem class.
    """

    def __init__(
        self,
        persistence: Optional[str] = "filesystem",
        root_dir: Optional[str] = None,
        auto_git_enabled: bool = False,
        composite_config: Optional[List[Dict[str, Any]]] = None,
        **storage_options,
    ):
        if composite_config is not None:
            self.configs = [
                Config(
                    persistence=config["persistence"],
                    root_dir=config["root_dir"],
                    is_auto_git_enabled=auto_git_enabled,
                    **storage_options,
                )
                for config in composite_config
            ]
        else:
            self.configs = [
                Config(
                    persistence=persistence,
                    root_dir=root_dir,
                    is_auto_git_enabled=auto_git_enabled,
                    **storage_options,
                ),
            ]

    @property
    def config(self):
        """
        Returns a single config.

        Exists to promote backwards compatibility.

        Returns
        -------
        Config
            A single Config
        """
        return self.configs[0]

    @property
    def repository(self):
        if len(self.configs) > 1:
            raise ValueError("More than one repository available. Use `.repositories` instead.")
        return self.configs[0].repository

    @property
    def repositories(self):
        return [config.repository for config in self.configs]

    @repository.setter
    def repository(self, value):
        if len(self.configs) > 1:
            raise ValueError("Cannot set when more than one repository available!")

        self.configs[0].repository = value

    def _get_github_url(self):
        """Returns the repository URL of the `git` repo it is called from."""
        completed_process = subprocess.run(["git", "remote", "-v"], capture_output=True)
        remotes = completed_process.stdout.decode("utf8").replace("\t", " ").split("\n")

        try:
            origin = [remote for remote in remotes if remote.startswith("origin")][0]
            github_url = origin.split(" ")[1]
        except IndexError:
            github_url = None

        return github_url

    def is_auto_git_enabled(self) -> bool:
        """Check if git is enabled for any of the configs."""
        if isinstance(self.configs, list):
            return any(_config.is_auto_git_enabled for _config in self.configs)

        if self.configs is None:
            return False

        return self.configs.is_auto_git_enabled

    def _create_project_domain(
        self,
        name: str,
        description: Optional[str],
        github_url: Optional[str],
        training_metadata: Optional[Union[List[Tuple], Tuple]],
    ):
        """Instantiates and returns a project domain object."""
        if self.is_auto_git_enabled() and github_url is None:
            github_url = self._get_github_url()

        if training_metadata is not None:
            training_metadata_class = TrainingMetadata(training_metadata)
        else:
            training_metadata_class = None

        return domain.Project(
            name,
            description=description,
            github_url=github_url,
            training_metadata=training_metadata_class,
        )

    @failsafe
    def create_project(
        self,
        name: str,
        description: Optional[str] = None,
        github_url: Optional[str] = None,
        training_metadata: Optional[Union[Tuple, List[Tuple]]] = None,
    ) -> Project:
        """Create a project.

        Parameters
        ----------
        name : str
            The project's name.
        description : str, optional
            The project's description.
        github_url : str, optional
            The URL of the GitHub repository associated with this
            project. If omitted and automatic `git` logging is
            enabled, it will be retrieved via `git remote`.
        training_metadata : tuple or list of tuples, optional
            Metadata associated with the training dataset(s)
            used across each experiment in this project.

        Returns
        -------
        rubicon.client.Project
            The created project.
        """
        project = self._create_project_domain(name, description, github_url, training_metadata)
        for repo in self.repositories:
            repo.create_project(project)

        return Project(project, self.configs)

    @failsafe
    def get_project(self, name: Optional[str] = None, id: Optional[str] = None) -> Project:
        """Get a project.

        Parameters
        ----------
        name : str, optional
            The name of the project to get.
        id : str, optional
            The id of the project to get.

        Returns
        -------
        rubicon.client.Project
            The project with name `name` or id `id`.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            return_err = None

            for repo in self.repositories:
                try:
                    project = repo.get_project(name)
                except Exception as err:
                    return_err = err
                else:
                    return Project(project, self.config)

            if len(self.repositories) > 1:
                raise RubiconException("all configured storage backends failed") from return_err
            else:
                raise return_err
        else:
            return [p for p in self.projects() if p.id == id][0]

    def get_project_as_dask_df(self, name, group_by=None):
        """DEPRECATED: Available for backwards compatibility."""
        warnings.warn(
            "`get_project_as_dask_df` is deprecated and will be removed in a future "
            "release. use `get_project_as_df('name', df_type='dask') instead.",
            DeprecationWarning,
        )

        return self.get_project_as_df(name, df_type="dask", group_by=group_by)

    @failsafe
    def get_project_as_df(self, name, df_type="pandas", group_by=None):
        """Get a dask or pandas dataframe representation of a project.

        Parameters
        ----------
        name : str
            The name of the project to get.
        df_type : str, optional
            The type of dataframe to return. Valid options include
            ["dask", "pandas"]. Defaults to "pandas".
        group_by : str or None, optional
            How to group the project's experiments in the returned
            DataFrame(s). Valid options include ["commit_hash"].

        Returns
        -------
        pandas.DataFrame or list of pandas.DataFrame or dask.DataFrame or list of dask.DataFrame
            If `group_by` is `None`, a dask or pandas dataframe holding the project's
            data. Otherwise a list of dask or pandas dataframes holding the project's
            data grouped by `group_by`.
        """
        project = self.get_project(name)

        return project.to_df(df_type=df_type, group_by=None)

    @failsafe
    def get_or_create_project(self, name: str, **kwargs) -> Project:
        """Get or create a project.

        Parameters
        ----------
        name : str
            The project's name.
        kwargs : dict
            Additional keyword arguments to be passed to
            `Rubicon.create_project`.

        Returns
        -------
        rubicon.client.Project
            The corresponding project.
        """
        try:
            project = self.get_project(name)
        except RubiconException:
            project = self.create_project(name, **kwargs)
        else:  # check for None in case of failure mode being set to "log" or "warn"
            if project is None:
                project = self.create_project(name, **kwargs)

        return project

    @failsafe
    def projects(self):
        """Get a list of available projects.

        Returns
        -------
        list of rubicon.client.Project
            The list of available projects.
        """
        return_err = None

        for repo in self.repositories:
            try:
                projects = [Project(project, self.config) for project in repo.get_projects()]
            except Exception as err:
                return_err = err
            else:
                return projects

        if len(self.repositories) > 1:
            raise RubiconException("all configured storage backends failed") from return_err
        else:
            raise return_err

    @failsafe
    def sync(
        self,
        project_name: str,
        s3_root_dir: str,
        aws_profile: Optional[str] = None,
        aws_shared_credentials_file: Optional[str] = None,
    ):
        """Sync a local project to S3.

        Parameters
        ----------
        project_name : str
            The name of the project to sync.
        s3_root_dir : str
            The S3 path where the project's data
            will be synced.
        aws_profile : str
            Specifies the name of the AWS CLI profile with the credentials and options to use.
            Defaults to None, using the AWS default name 'default'.
        aws_shared_credentials_file : str
            Specifies the location of the file that the AWS CLI uses to store access keys.
            Defaults to None, using the AWS default path '~/.aws/credentials'.

        Notes
        -----
        Use sync to backup your local project data to S3 as an alternative to direct S3 logging.
        Leverages the AWS CLI's `aws s3 sync`. Ensure that any credentials are set and that any
        proxies are enabled.
        """
        if self.config.persistence != "filesystem":
            raise RubiconException("Projects can only be synced from local or S3 filesystems.")

        cmd_root = "aws s3 sync"

        if aws_profile:
            cmd_root += f" --profile {aws_profile}"

        original_aws_shared_credentials_file = os.environ.get("AWS_SHARED_CREDENTIALS_FILE")

        if aws_shared_credentials_file:
            os.environ["AWS_SHARED_CREDENTIALS_FILE"] = aws_shared_credentials_file

        project = self.get_project(project_name)
        local_path = f"{self.config.root_dir}/{slugify(project.name)}"
        cmd = f"{cmd_root} {local_path} {s3_root_dir}/{slugify(project.name)}"

        try:
            subprocess.run(cmd, shell=True, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            raise RubiconException(e.stderr)
        finally:
            if aws_shared_credentials_file:
                if original_aws_shared_credentials_file:
                    os.environ["AWS_SHARED_CREDENTIALS_FILE"] = original_aws_shared_credentials_file
                else:
                    del os.environ["AWS_SHARED_CREDENTIALS_FILE"]



from __future__ import annotations

from typing import TYPE_CHECKING, List

from rubicon_ml import domain
from rubicon_ml.client import (
    ArtifactMixin,
    Base,
    CommentMixin,
    DataframeMixin,
    Feature,
    Metric,
    Parameter,
    TagMixin,
)
from rubicon_ml.client.utils.exception_handling import failsafe
from rubicon_ml.client.utils.tags import filter_children
from rubicon_ml.exceptions import RubiconException

if TYPE_CHECKING:
    from rubicon_ml.client import Project
    from rubicon_ml.domain import Experiment as ExperimentDomain


class Experiment(Base, ArtifactMixin, DataframeMixin, TagMixin, CommentMixin):
    """A client experiment.

    An `experiment` represents a model run and is identified by
    its 'created_at' time. It can have `metrics`, `parameters`,
    `features`, `dataframes`, and `artifacts` logged to it.

    An `experiment` is logged to a `project`.

    Parameters
    ----------
    domain : rubicon.domain.Experiment
        The experiment domain model.
    parent : rubicon.client.Project
        The project that the experiment is logged to.
    """

    def __init__(self, domain: ExperimentDomain, parent: Project):
        super().__init__(domain, parent._config)

        self._domain: ExperimentDomain

        self._parent = parent
        self._artifacts = []
        self._dataframes = []
        self._metrics = []
        self._features = []
        self._parameters = []

    def _get_identifiers(self):
        """Get the experiment's project's name and the experiment's ID."""
        return self.project.name, self.id

    @failsafe
    def log_metric(
        self,
        name: str,
        value: float,
        directionality: str = "score",
        description: str = None,
        tags: list[str] = [],
        comments: list[str] = [],
    ) -> Metric:
        """Create a metric under the experiment.

        Parameters
        ----------
        name : str
            The metric's name.
        value : float
            The metric's value.
        directionality : str, optional
            The metric's directionality. Must be one of
            ["score", "loss"], where "score" represents
            a metric to maximize, while "loss" represents a
            metric to minimize. Defaults to "score".
        description : str, optional
            The metric's description. Use to provide additional
            context.
        tags : list of str, optional
            Values to tag the experiment with. Use tags to organize and
            filter your metrics.
        comments : list of str, optional
            Values to comment the experiment with. Use comments to organize and
            filter your metrics.

        Returns
        -------
        rubicon.client.Metric
            The created metric.
        """
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        if not isinstance(comments, list) or not all(
            [isinstance(comment, str) for comment in comments]
        ):
            raise ValueError("`comments` must be `list` of type `str`")

        metric = domain.Metric(
            name,
            value,
            directionality=directionality,
            description=description,
            tags=tags,
            comments=comments,
        )
        for repo in self.repositories:
            repo.create_metric(metric, self.project.name, self.id)

        return Metric(metric, self)

    @failsafe
    def metrics(self, name=None, tags=[], qtype="or"):
        """Get the metrics logged to this experiment.

        Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Metric
            The metrics previously logged to this experiment.
        """
        return_err = None

        for repo in self.repositories:
            try:
                metrics = [Metric(m, self) for m in repo.get_metrics(self.project.name, self.id)]
            except Exception as err:
                return_err = err
            else:
                self._metrics = filter_children(metrics, tags, qtype, name)

                return self._metrics

        self._raise_rubicon_exception(return_err)

    @failsafe
    def metric(self, name=None, id=None):
        """Get a metric.

        Parameters
        ----------
        name : str, optional
            The name of the metric to get.
        id : str, optional
            The id of the metric to get.

        Returns
        -------
        rubicon.client.Metric
            The metric with name `name` or id `id`.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            return_err = None

            for repo in self.repositories:
                try:
                    metric = repo.get_metric(self.project.name, self.id, name)
                except Exception as err:
                    return_err = err
                else:
                    return Metric(metric, self)

            self._raise_rubicon_exception(return_err)
        else:
            return [m for m in self.metrics() if m.id == id][0]

    @failsafe
    def log_feature(
        self,
        name: str,
        description: str = None,
        importance: float = None,
        tags: list[str] = [],
        comments: list[str] = [],
    ) -> Feature:
        """Create a feature under the experiment.

        Parameters
        ----------
        name : str
            The features's name.
        description : str
            The feature's description. Use to provide
            additional context.
        importance : float
            The feature's importance.
        tags : list of str, optional
            Values to tag the experiment with. Use tags to organize and
            filter your features.
        comments : list of str, optional
            Values to comment the experiment with. Use comments to organize and
            filter your features.

        Returns
        -------
        rubicon.client.Feature
            The created feature.
        """
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        if not isinstance(comments, list) or not all(
            [isinstance(comment, str) for comment in comments]
        ):
            raise ValueError("`comments` must be `list` of type `str`")

        feature = domain.Feature(
            name, description=description, importance=importance, tags=tags, comments=comments
        )

        for repo in self.repositories:
            repo.create_feature(feature, self.project.name, self.id)

        return Feature(feature, self)

    @failsafe
    def features(self, name=None, tags=[], qtype="or"):
        """Get the features logged to this experiment.

         Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Feature
            The features previously logged to this experiment.
        """
        return_err = None

        for repo in self.repositories:
            try:
                features = [Feature(f, self) for f in repo.get_features(self.project.name, self.id)]
            except Exception as err:
                return_err = err
            else:
                self._features = filter_children(features, tags, qtype, name)

                return self._features

        self._raise_rubicon_exception(return_err)

    @failsafe
    def feature(self, name=None, id=None):
        """Get a feature.

        Parameters
        ----------
        name : str, optional
            The name of the feature to get.
        id : str, optional
            The id of the feature to get.

        Returns
        -------
        rubicon.client.Feature
            The feature with name `name` or id `id`.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            return_err = None

            for repo in self.repositories:
                try:
                    feature = repo.get_feature(self.project.name, self.id, name)
                except Exception as err:
                    return_err = err
                else:
                    return Feature(feature, self)

            self._raise_rubicon_exception(return_err)
        else:
            return [f for f in self.features() if f.id == id][0]

    @failsafe
    def log_parameter(
        self,
        name: str,
        value: object = None,
        description: str = None,
        tags: list[str] = [],
        comments: list[str] = [],
    ) -> Parameter:
        """Create a parameter under the experiment.

        Parameters
        ----------
        name : str
            The parameter's name.
        value : object, optional
            The parameter's value. Can be an object of any JSON
            serializable (via rubicon.utils.DomainJSONEncoder)
            type.
        description : str, optional
            The parameter's description. Use to provide
            additional context.
        tags : list of str, optional
            Values to tag the parameter with. Use tags to organize and
            filter your parameters.
        comments : list of str, optional
            Values to comment the experiment with. Use comments to organize and
            filter your features.

        Returns
        -------
        rubicon.client.Parameter
            The created parameter.
        """
        if not isinstance(tags, list) or not all([isinstance(tag, str) for tag in tags]):
            raise ValueError("`tags` must be `list` of type `str`")

        if not isinstance(comments, list) or not all(
            [isinstance(comment, str) for comment in comments]
        ):
            raise ValueError("`comments` must be `list` of type `str`")

        parameter = domain.Parameter(
            name, value=value, description=description, tags=tags, comments=comments
        )

        for repo in self.repositories:
            repo.create_parameter(parameter, self.project.name, self.id)

        return Parameter(parameter, self)

    @failsafe
    def parameters(self, name=None, tags=[], qtype="or"):
        """Get the parameters logged to this experiment.

        Parameters
        ----------
        name : str, optional
            The name value to filter results on.
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Parameter
            The parameters previously logged to this experiment.
        """
        return_err = None

        for repo in self.repositories:
            try:
                parameters = [
                    Parameter(p, self) for p in repo.get_parameters(self.project.name, self.id)
                ]
            except Exception as err:
                return_err = err
            else:
                self._parameters = filter_children(parameters, tags, qtype, name)

                return self._parameters

        self._raise_rubicon_exception(return_err)

    @failsafe
    def parameter(self, name=None, id=None):
        """Get a parameter.

        Parameters
        ----------
        name : str, optional
            The name of the parameter to get.
        id : str, optional
            The id of the parameter to get.

        Returns
        -------
        rubicon.client.Parameter
            The parameter with name `name` or id `id`.
        """
        if (name is None and id is None) or (name is not None and id is not None):
            raise ValueError("`name` OR `id` required.")

        if name is not None:
            return_err = None

            for repo in self.repositories:
                try:
                    parameter = repo.get_parameter(self.project.name, self.id, name)
                except Exception as err:
                    return_err = err
                else:
                    return Parameter(parameter, self)

            self._raise_rubicon_exception(return_err)
        else:
            return [p for p in self.parameters() if p.id == id][0]

    def add_child_experiment(self, experiment: Experiment):
        """Add tags to denote `experiment` as a descendent of this experiment.

        Parameters
        ----------
        experiment : rubicon_ml.client.Experiment
            The experiment to mark as a descendent of this experiment.

        Raises
        ------
        RubiconException
            If `experiment` and this experiment are not logged to the same project.
        """
        if experiment.project.id != self.project.id:
            raise RubiconException(
                "Descendents must be logged to the same project. Project"
                f"{experiment.project.id} does not match project {self.project.id}."
            )

        child_tag = f"child:{experiment.id}"
        parent_tag = f"parent:{self.id}"

        self.add_tags([child_tag])
        experiment.add_tags([parent_tag])

    def _get_experiments_from_tags(self, tag_key: str):
        """Get the experiments with `experiment_id`s in this experiment's tags
        that match the format `tag_key:experiment_id`.

        Returns
        -------
        list of rubicon_ml.client.Experiment
            The experiments with `experiment_id`s in this experiment's tags.
        """
        try:
            experiment_ids = self.tags[tag_key]
        except KeyError:
            return []

        if not isinstance(experiment_ids, list):
            experiment_ids = [experiment_ids]

        return [self.project.experiment(id=exp_id) for exp_id in experiment_ids]

    def get_child_experiments(self) -> List[Experiment]:
        """Get the experiments that are tagged as children of this experiment.

        Returns
        -------
        list of rubicon_ml.client.Experiment
            The experiments that are tagged as children of this experiment.
        """
        return self._get_experiments_from_tags("child")

    def get_parent_experiments(self) -> List[Experiment]:
        """Get the experiments that are tagged as parents of this experiment.

        Returns
        -------
        list of rubicon_ml.client.Experiment
            The experiments that are tagged as parents of this experiment.
        """
        return self._get_experiments_from_tags("parent")

    @property
    def id(self):
        """Get the experiment's id."""
        return self._domain.id

    @property
    def name(self):
        """Get the experiment's name."""
        return self._domain.name

    @property
    def description(self):
        """Get the experiment's description."""
        return self._domain.description

    @property
    def model_name(self):
        """Get the experiment's model name."""
        return self._domain.model_name

    @property
    def branch_name(self):
        """Get the experiment's branch name."""
        return self._domain.branch_name

    @property
    def commit_hash(self):
        """Get the experiment's commit hash."""
        return self._domain.commit_hash

    @property
    def training_metadata(self):
        """Get the project's training metadata."""
        training_metadata = self._domain.training_metadata.training_metadata

        if len(training_metadata) == 1:
            training_metadata = training_metadata[0]

        return training_metadata

    @property
    def created_at(self):
        """Get the time the experiment was created."""
        return self._domain.created_at

    @property
    def project(self):
        """Get the project client object that this experiment
        belongs to.
        """
        return self._parent



import json

import numpy as np
import plotly.graph_objects as go
from dash import callback_context, dcc, html
from dash.dependencies import ALL, Input, Output

from rubicon_ml.viz.base import VizBase
from rubicon_ml.viz.common import dropdown_header
from rubicon_ml.viz.common.colors import get_rubicon_colorscale, light_blue, transparent


class MetricCorrelationPlot(VizBase):
    """Visualize the correlation between the parameters and metrics logged
    to the experiments `experiments` using a parallel coordinates plot.

    More info on parallel coordinates plots can be found here:
    https://plotly.com/python/parallel-coordinates-plot/

    Parameters
    ----------
    experiments : list of rubicon_ml.client.experiment.Experiment, optional
        The experiments to visualize. Defaults to None. Can be set as
        attribute after instantiation.
    metric_names : list of str
        The names of the metrics to load. Defaults to None, which loads all
        metrics logged to the experiments `experiments`.
    parameter_names : list of str
        The names of the parameters to load. Defaults to None, which loads all
        parameters logged to the experiments `experiments`.
    selected_metric : str
        The name of the metric to display at launch. Defaults to None, which
        selects the metric loaded first.
    """

    def __init__(
        self,
        experiments=None,
        metric_names=None,
        parameter_names=None,
        selected_metric=None,
    ):
        super().__init__(dash_title="plot metric correlation")

        self.experiments = experiments
        self.metric_names = metric_names
        self.parameter_names = parameter_names
        self.selected_metric = selected_metric

    def _get_dimension(self, label, values):
        """Transforms the input data for use with Plotly's parallel
        coordinates plot.
        """
        if len(values) > 0 and any([isinstance(v, str) or isinstance(v, bool) for v in values]):
            values = [str(v) for v in values]
            unique_values, values = np.unique(values, return_inverse=True)

            dimension = {
                "label": label,
                "ticktext": unique_values,
                "tickvals": list(range(0, len(unique_values))),
                "values": values,
            }
        else:
            dimension = {
                "label": label,
                "values": values,
            }

        return dimension

    @property
    def layout(self):
        """Defines the layout for the metric correlation plot."""
        return html.Div(
            [
                dropdown_header(
                    self.visible_metric_names,
                    self.selected_metric,
                    "comparing metric ",
                    f" over {len(self.experiments)} experiments",
                    "metric-correlation",
                ),
                dcc.Loading(
                    html.Div(
                        dcc.Graph(
                            id="metric-correlation-plot",
                        ),
                        id="metric-correlation-plot-container",
                    ),
                    color=light_blue,
                ),
            ],
            id="metric-correlation-plot-layout-container",
        )

    def load_experiment_data(self):
        """Load the experiment data required for the experiments table.

        Extracts parameter and metric metadata from each experiment in
        `self.experiments`. List metrics are ignored.
        """
        self.experiment_records = {}
        self.visible_metric_names = set()
        self.visible_parameter_names = set()

        for experiment in self.experiments:
            experiment_record = {"metrics": {}, "parameters": {}}

            for metric in experiment.metrics():
                if (
                    self.metric_names is None or metric.name in self.metric_names
                ) and not isinstance(metric.value, list):
                    experiment_record["metrics"][metric.name] = metric.value

                    self.visible_metric_names.add(metric.name)

                    if self.selected_metric is None:
                        self.selected_metric = metric.name

            for parameter in experiment.parameters():
                if self.parameter_names is None or parameter.name in self.parameter_names:
                    experiment_record["parameters"][parameter.name] = parameter.value

                    self.visible_parameter_names.add(parameter.name)

            self.experiment_records[experiment.id] = experiment_record

        if self.selected_metric not in self.visible_metric_names:
            raise ValueError(
                f"no metric named `selected_metric` '{self.selected_metric}'"
                " logged to any experiment in `experiments`."
            )

        self.visible_parameter_names = list(self.visible_parameter_names)
        self.visible_metric_names = list(self.visible_metric_names)
        self.visible_metric_names.sort()

    def register_callbacks(self, link_experiment_table=False):
        outputs = [
            Output("metric-correlation-plot", "figure"),
            Output("metric-correlation-dropdown", "label"),
            Output("metric-correlation-header-right-text", "children"),
        ]
        inputs = [
            Input({"type": "metric-correlation-dropdown-button", "index": ALL}, "n_clicks"),
        ]
        states = []

        if link_experiment_table:
            inputs.append(
                Input("experiment-table", "derived_virtual_selected_row_ids"),
            )

        @self.app.callback(outputs, inputs, states)
        def update_metric_plot(*args):
            """Render the correlation plot based on the currently selected metric.

            Returns the Plotly `Parcoords` figure generated by the values of the
            experiments' selected metric, the name of the currently selected
            metric, and the header text with the metric's name.
            """
            if link_experiment_table:
                selected_row_ids = args[-1]
                selected_row_ids = selected_row_ids if selected_row_ids else []

                experiment_records = [
                    self.experiment_records[row_id] for row_id in selected_row_ids
                ]
            else:
                experiment_records = self.experiment_records.values()

            property_id = callback_context.triggered[0].get("prop_id")
            property_value = property_id[: property_id.index(".")]

            if not property_value or property_value == "experiment-table":
                selected_metric = self.selected_metric
            else:
                selected_metric = json.loads(property_value).get("index")

                self.selected_metric = selected_metric

            header_right_text = (
                f"over {len(experiment_records)} experiment"
                f"{'s' if len(experiment_records) != 1 else ''}"
            )

            parameter_values = {}

            for parameter_name in self.visible_parameter_names:
                parameter_values[parameter_name] = [
                    record["parameters"].get(parameter_name) for record in experiment_records
                ]

            metric_values = [
                record["metrics"].get(selected_metric) for record in experiment_records
            ]

            plot_dimensions = []

            for parameter_name, parameter_value in parameter_values.items():
                if any([p is not None for p in parameter_value]):
                    plot_dimensions.append(self._get_dimension(parameter_name, parameter_value))

            plot_dimensions.append(self._get_dimension(selected_metric, metric_values))

            metric_correlation_plot = go.Figure(
                go.Parcoords(
                    line={
                        "color": [m for m in metric_values if m is not None],
                        "colorscale": get_rubicon_colorscale(len(experiment_records)),
                        "showscale": True,
                    },
                    dimensions=plot_dimensions,
                ),
            )
            metric_correlation_plot.update_layout(paper_bgcolor=transparent)

            return metric_correlation_plot, selected_metric, header_right_text



from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional

from rubicon_ml.domain.mixin import CommentMixin, InitMixin, TagMixin

if TYPE_CHECKING:
    from datetime import datetime

    from rubicon_ml.domain.utils import TrainingMetadata


@dataclass(init=False)
class Experiment(CommentMixin, InitMixin, TagMixin):
    """A domain-level experiment.

    project_name : str
        The name of the project the experiment belongs to.
    branch_name : str, optional
        The name of the git branch associated with the experiment. Defaults to `None`.
    comments : list of str, optional
        Additional text information and observations about the experiment. Defaults to
        `None`.
    commit_hash : str, optional
        The hash of the git commit associated with the experiment. Defaults to `None`.
    created_at : datetime, optional
        The date and time the experiment was created. Defaults to `None` and uses
        `datetime.datetime.now` to generate a UTC timestamp. `created_at` should be
        left as `None` to allow for automatic generation.
    description : str, optional
        A description of the experiment. Defaults to `None`.
    id : str, optional
        The experiment's unique identifier. Defaults to `None` and uses `uuid.uuid4`
        to generate a unique ID. `id` should be left as `None` to allow for automatic
        generation.
    model_name : str, optional
        The name of the model associated with the experiment. Defaults to `None`.
    name : str, optional
        The experiment's name. Defaults to `None`.
    tags : list of str, optional
        The values this experiment is tagged with. Defaults to `None`.
    training_metadata : rubicon_ml.domain.utils.TrainingMetadata, optional
        Additional metadata pertaining to any data this experiment was trained on.
        Defaults to `None`.
    """

    project_name: str
    branch_name: Optional[str] = None
    comments: Optional[List[str]] = None
    commit_hash: Optional[str] = None
    created_at: Optional["datetime"] = None
    description: Optional[str] = None
    id: Optional[str] = None
    model_name: Optional[str] = None
    name: Optional[str] = None
    tags: Optional[List[str]] = None
    training_metadata: Optional["TrainingMetadata"] = None

    def __init__(
        self,
        project_name: str,
        branch_name: Optional[str] = None,
        comments: Optional[List[str]] = None,
        commit_hash: Optional[str] = None,
        created_at: Optional["datetime"] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        model_name: Optional[str] = None,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
        training_metadata: Optional["TrainingMetadata"] = None,
        **kwargs,
    ):
        """Initialize this domain experiment."""

        self._check_extra_kwargs(kwargs)

        self.project_name = project_name
        self.branch_name = branch_name
        self.comments = comments or []
        self.commit_hash = commit_hash
        self.created_at = self._init_created_at(created_at)
        self.description = description
        self.id = self._init_id(id)
        self.model_name = model_name
        self.name = name
        self.tags = tags or []
        self.training_metadata = training_metadata



from rubicon_ml.exceptions import RubiconException
from rubicon_ml.sklearn.estimator_logger import EstimatorLogger
from rubicon_ml.sklearn.utils import log_parameter_with_warning


class FilterEstimatorLogger(EstimatorLogger):
    """The filter logger for sklearn estimators. Use this
    logger to either select or ignore specific parameters
    for logging.

    Parameters
    ----------
    estimator : a sklearn estimator, optional
        The estimator
    experiment : rubicon.client.Experiment, optional
        The experiment to log the parameters and metrics to.
    step_name : str, optional
        The name of the pipeline step.
    select : list, optional
        The list of parameters on this estimator that you'd like
        to log. All other parameters will be ignored.
    ignore : list, optional
        The list of parameters on this estimator that you'd like
        to ignore by not logging. The other parameters will be logged.
    ignore_all : bool, optional
        Ignore all parameters if true.
    """

    def __init__(
        self,
        estimator=None,
        experiment=None,
        step_name=None,
        select=[],
        ignore=[],
        ignore_all=False,
    ):
        if ignore and select:
            raise RubiconException("provide either `select` OR `ignore`, not both")

        self.ignore = ignore
        self.ignore_all = ignore_all
        self.select = select

        super().__init__(estimator=estimator, experiment=experiment, step_name=step_name)

    def log_parameters(self):
        if self.ignore_all:
            return

        for name, value in self.estimator.get_params().items():
            if (self.ignore and name not in self.ignore) or (self.select and name in self.select):
                log_parameter_with_warning(self.experiment, f"{self.step_name}__{name}", value)
